@@ -124,7 +124,7 @@ read_pid_stat (pid_t pid, struct pid_stat *st, libcrun_error_t *err)
           memset (st, 0, sizeof (*st));
           return 0;
         }
-      return crun_make_error (err, errno, "open state file %s", pid_stat_file);
+      return crun_make_error (err, errno, "open state file `%s`", pid_stat_file);
     }
 
   ret = read_all_fd (fd, pid_stat_file, &buffer, NULL, err);
@@ -237,6 +237,15 @@ libcrun_write_container_status (const char *state_root, const char *id, libcrun_
   if (UNLIKELY (r != yajl_gen_status_ok))
     goto yajl_error;
 
+  r = yajl_gen_string (gen, YAJL_STR ("intelrdt"), strlen ("intelrdt"));
+  if (UNLIKELY (r != yajl_gen_status_ok))
+    goto yajl_error;
+
+  tmp = status->intelrdt ? status->intelrdt : "";
+  r = yajl_gen_string (gen, YAJL_STR (tmp), strlen (tmp));
+  if (UNLIKELY (r != yajl_gen_status_ok))
+    goto yajl_error;
+
   r = yajl_gen_string (gen, YAJL_STR ("rootfs"), strlen ("rootfs"));
   if (UNLIKELY (r != yajl_gen_status_ok))
     goto yajl_error;
@@ -269,6 +278,17 @@ libcrun_write_container_status (const char *state_root, const char *id, libcrun_
   if (UNLIKELY (r != yajl_gen_status_ok))
     goto yajl_error;
 
+  if (status->owner)
+    {
+      r = yajl_gen_string (gen, YAJL_STR ("owner"), strlen ("owner"));
+      if (UNLIKELY (r != yajl_gen_status_ok))
+        goto yajl_error;
+
+      r = yajl_gen_string (gen, YAJL_STR (status->owner), strlen (status->owner));
+      if (UNLIKELY (r != yajl_gen_status_ok))
+        goto yajl_error;
+    }
+
   r = yajl_gen_string (gen, YAJL_STR ("detached"), strlen ("detached"));
   if (UNLIKELY (r != yajl_gen_status_ok))
     goto yajl_error;
@@ -336,13 +356,13 @@ libcrun_read_container_status (libcrun_container_status_t *status, const char *s
 
   tree = yajl_tree_parse (buffer, err_buffer, sizeof (err_buffer));
   if (UNLIKELY (tree == NULL))
-    return crun_make_error (err, 0, "cannot parse status file: %s", err_buffer);
+    return crun_make_error (err, 0, "cannot parse status file: `%s`", err_buffer);
 
   {
     const char *pid_path[] = { "pid", NULL };
     tmp = yajl_tree_get (tree, pid_path, yajl_t_number);
     if (UNLIKELY (tmp == NULL))
-      return crun_make_error (err, 0, "'pid' missing in %s", file);
+      return crun_make_error (err, 0, "`pid` missing in `%s`", file);
     status->pid = strtoull (YAJL_GET_NUMBER (tmp), NULL, 10);
   }
   {
@@ -357,7 +377,7 @@ libcrun_read_container_status (libcrun_container_status_t *status, const char *s
     const char *cgroup_path[] = { "cgroup-path", NULL };
     tmp = yajl_tree_get (tree, cgroup_path, yajl_t_string);
     if (UNLIKELY (tmp == NULL))
-      return crun_make_error (err, 0, "'cgroup-path' missing in %s", file);
+      return crun_make_error (err, 0, "`cgroup-path` missing in `%s`", file);
     status->cgroup_path = xstrdup (YAJL_GET_STRING (tmp));
   }
   {
@@ -365,11 +385,16 @@ libcrun_read_container_status (libcrun_container_status_t *status, const char *s
     tmp = yajl_tree_get (tree, scope, yajl_t_string);
     status->scope = tmp ? xstrdup (YAJL_GET_STRING (tmp)) : NULL;
   }
+  {
+    const char *intelrdt[] = { "intelrdt", NULL };
+    tmp = yajl_tree_get (tree, intelrdt, yajl_t_string);
+    status->intelrdt = tmp ? xstrdup (YAJL_GET_STRING (tmp)) : NULL;
+  }
   {
     const char *rootfs[] = { "rootfs", NULL };
     tmp = yajl_tree_get (tree, rootfs, yajl_t_string);
     if (UNLIKELY (tmp == NULL))
-      return crun_make_error (err, 0, "'rootfs' missing in %s", file);
+      return crun_make_error (err, 0, "`rootfs` missing in `%s`", file);
     status->rootfs = xstrdup (YAJL_GET_STRING (tmp));
   }
   {
@@ -380,51 +405,29 @@ libcrun_read_container_status (libcrun_container_status_t *status, const char *s
     const char *bundle[] = { "bundle", NULL };
     tmp = yajl_tree_get (tree, bundle, yajl_t_string);
     if (UNLIKELY (tmp == NULL))
-      return crun_make_error (err, 0, "'bundle' missing in %s", file);
+      return crun_make_error (err, 0, "`bundle` missing in `%s`", file);
     status->bundle = xstrdup (YAJL_GET_STRING (tmp));
   }
   {
     const char *created[] = { "created", NULL };
     tmp = yajl_tree_get (tree, created, yajl_t_string);
     if (UNLIKELY (tmp == NULL))
-      return crun_make_error (err, 0, "'created' missing in %s", file);
+      return crun_make_error (err, 0, "`created` missing in `%s`", file);
     status->created = xstrdup (YAJL_GET_STRING (tmp));
   }
+  {
+    const char *owner[] = { "owner", NULL };
+    tmp = yajl_tree_get (tree, owner, yajl_t_string);
+    status->owner = tmp ? xstrdup (YAJL_GET_STRING (tmp)) : NULL;
+  }
   {
     const char *detached[] = { "detached", NULL };
     status->detached = YAJL_IS_TRUE (yajl_tree_get (tree, detached, yajl_t_true));
   }
   {
-    const char *external[] = { "external_descriptors", NULL };
-    const unsigned char *buf = NULL;
-    yajl_gen gen = NULL;
-    size_t buf_len;
-
-    gen = yajl_gen_alloc (NULL);
-    if (gen == NULL)
-      return crun_make_error (err, errno, "yajl_gen_alloc");
-    yajl_gen_array_open (gen);
-
-    tmp = yajl_tree_get (tree, external, yajl_t_array);
-    if (tmp && YAJL_IS_ARRAY (tmp))
-      {
-        size_t len = tmp->u.array.len;
-        size_t i;
-        for (i = 0; i < len; ++i)
-          {
-            yajl_val s = tmp->u.array.values[i];
-            if (s && YAJL_IS_STRING (s))
-              {
-                char *str = YAJL_GET_STRING (s);
-                yajl_gen_string (gen, YAJL_STR (str), strlen (str));
-              }
-          }
-      }
-    yajl_gen_array_close (gen);
-    yajl_gen_get_buf (gen, &buf, &buf_len);
-    if (buf)
-      status->external_descriptors = xstrdup ((const char *) buf);
-    yajl_gen_free (gen);
+    const char *external_descriptors[] = { "external_descriptors", NULL };
+    tmp = yajl_tree_get (tree, external_descriptors, yajl_t_string);
+    status->external_descriptors = tmp ? xstrdup (YAJL_GET_STRING (tmp)) : NULL;
   }
   yajl_tree_free (tree);
   return 0;
@@ -464,7 +467,7 @@ rmdirfd (const char *namedir, int fd, libcrun_error_t *err)
   int ret;
   cleanup_dir DIR *d = NULL;
   struct dirent *de;
-  cleanup_close int fd_cleanup = fd;
+  __attribute__ ((unused)) cleanup_close int fd_cleanup = fd;
 
   d = fdopendir (fd);
   if (d == NULL)
@@ -521,7 +524,7 @@ libcrun_container_delete_status (const char *state_root, const char *id, libcrun
 
   dfd = openat (rundir_dfd, id, O_DIRECTORY | O_RDONLY);
   if (UNLIKELY (dfd < 0))
-    return crun_make_error (err, errno, "cannot open directory '%s/%s'", dir, id);
+    return crun_make_error (err, errno, "cannot open directory `%s/%s`", dir, id);
 
   ret = rmdirfd (dir, dfd, err);
 
@@ -549,6 +552,8 @@ libcrun_free_container_status (libcrun_container_status_t *status)
   free (status->external_descriptors);
   free (status->created);
   free (status->scope);
+  free (status->intelrdt);
+  free (status->owner);
 }
 
 int
@@ -586,7 +591,10 @@ libcrun_get_containers_list (libcrun_container_list_t **ret, const char *state_r
         }
 
       if (! exists)
-        continue;
+        {
+          libcrun_error (errno, "error opening file `%s`", status_file);
+          continue;
+        }
 
       next_container = xmalloc (sizeof (libcrun_container_list_t));
       next_container->name = xstrdup (next->d_name);
@@ -699,7 +707,7 @@ libcrun_status_write_exec_fifo (const char *state_root, const char *id, libcrun_
 
   ret = TEMP_FAILURE_RETRY (write (fd, buffer, 1));
   if (UNLIKELY (ret < 0))
-    return crun_make_error (err, errno, "read from exec.fifo");
+    return crun_make_error (err, errno, "write to exec.fifo");
 
   return strtoll (buffer, NULL, 10);
 }
