@@ -94,7 +94,7 @@ crun_path_exists (const char *path, libcrun_error_t *err arg_unused)
 {
   int ret = access (path, F_OK);
   if (ret < 0)
-    return 0;
+    return errno == ENOENT ? 0 : ret;
   return 1;
 }
 
@@ -205,6 +205,7 @@ fallback:
 int
 get_file_type_at (int dirfd, mode_t *mode, bool nofollow, const char *path)
 {
+  int empty_path = path == NULL ? AT_EMPTY_PATH : 0;
   struct stat st;
   int ret;
 
@@ -213,7 +214,7 @@ get_file_type_at (int dirfd, mode_t *mode, bool nofollow, const char *path)
     0,
   };
 
-  ret = statx (dirfd, path, (nofollow ? AT_SYMLINK_NOFOLLOW : 0) | AT_STATX_DONT_SYNC, STATX_TYPE, &stx);
+  ret = statx (dirfd, path ?: "", empty_path | (nofollow ? AT_SYMLINK_NOFOLLOW : 0) | AT_STATX_DONT_SYNC, STATX_TYPE, &stx);
   if (UNLIKELY (ret < 0))
     {
       if (errno == ENOSYS || errno == EINVAL)
@@ -226,7 +227,7 @@ get_file_type_at (int dirfd, mode_t *mode, bool nofollow, const char *path)
 
 fallback:
 #endif
-  ret = fstatat (dirfd, path, &st, nofollow ? AT_SYMLINK_NOFOLLOW : 0);
+  ret = fstatat (dirfd, path ?: "", &st, empty_path | (nofollow ? AT_SYMLINK_NOFOLLOW : 0));
   *mode = st.st_mode;
   return ret;
 }
@@ -333,7 +334,7 @@ check_fd_under_path (const char *rootfs, size_t rootfslen, int fd, const char *f
   if (UNLIKELY (ret < 0))
     return crun_make_error (err, errno, "readlink `%s`", fdname);
 
-  if (((size_t) ret) <= rootfslen || memcmp (link, rootfs, rootfslen) != 0)
+  if (((size_t) ret) <= rootfslen || memcmp (link, rootfs, rootfslen) != 0 || link[rootfslen] != '/')
     return crun_make_error (err, 0, "target `%s` not under the directory `%s`", fdname, rootfs);
 
   return 0;
@@ -421,7 +422,7 @@ safe_openat (int dirfd, const char *rootfs, size_t rootfs_len, const char *path,
   return safe_openat_fallback (dirfd, rootfs, rootfs_len, path, flags, mode, err);
 }
 
-static ssize_t
+ssize_t
 safe_readlinkat (int dfd, const char *name, char **buffer, ssize_t hint, libcrun_error_t *err)
 {
   ssize_t buf_size = hint > 0 ? hint + 1 : 512;
@@ -546,9 +547,28 @@ crun_safe_ensure_at (bool do_open, bool dir, int dirfd, const char *dirpath,
             return crun_make_error (err, errno, "mkdir `/%s`", npath);
         }
 
-      cwd = safe_openat (dirfd, dirpath, dirpath_len, npath, O_CLOEXEC | O_PATH, 0, err);
+      cwd = safe_openat (dirfd, dirpath, dirpath_len, npath, (last_component ? O_PATH : 0) | O_CLOEXEC, 0, err);
       if (UNLIKELY (cwd < 0))
-        return cwd;
+        return crun_error_wrap (err, "creating `/%s`", path);
+
+      if (! last_component)
+        {
+          mode_t st_mode;
+
+          ret = get_file_type_at (cwd, &st_mode, true, NULL);
+          if (UNLIKELY (ret < 0))
+            {
+              int saved_errno = errno;
+
+              close (cwd);
+              return crun_make_error (err, saved_errno, "error stat'ing file `%s`", npath);
+            }
+          if ((st_mode & S_IFMT) != S_IFDIR)
+            {
+              close (cwd);
+              return crun_make_error (err, ENOTDIR, "error creating directory `%s` since `%s` exists and it is not a directory", path, npath);
+            }
+        }
 
       close_and_replace (&wd_cleanup, cwd);
 
@@ -736,7 +756,7 @@ libcrun_initialize_selinux (libcrun_error_t *err)
 
   fd = open ("/proc/mounts", O_RDONLY | O_CLOEXEC);
   if (UNLIKELY (fd < 0))
-    return crun_make_error (err, errno, "open /proc/mounts");
+    return crun_make_error (err, errno, "open `/proc/mounts`");
 
   ret = read_all_fd_with_size_hint (fd, "/proc/mounts", &out, &len, get_page_size (), err);
   if (UNLIKELY (ret < 0))
@@ -780,7 +800,7 @@ libcrun_is_selinux_enabled (libcrun_error_t *err)
 }
 
 int
-add_selinux_mount_label (char **retlabel, const char *data, const char *label, const char *context_type, libcrun_error_t *err arg_unused)
+add_selinux_mount_label (char **retlabel, const char *data, const char *label, const char *context_type, libcrun_error_t *err)
 {
   int ret;
 
@@ -1164,7 +1184,7 @@ epoll_helper (int *fds, int *levelfds, libcrun_error_t *err)
       ev.data.fd = *it;
       ret = epoll_ctl (epollfd, EPOLL_CTL_ADD, *it, &ev);
       if (UNLIKELY (ret < 0))
-        return crun_make_error (err, errno, "epoll_ctl add '%d'", *it);
+        return crun_make_error (err, errno, "epoll_ctl add `%d`", *it);
     }
   for (it = levelfds; *it >= 0; it++)
     {
@@ -1172,7 +1192,7 @@ epoll_helper (int *fds, int *levelfds, libcrun_error_t *err)
       ev.data.fd = *it;
       ret = epoll_ctl (epollfd, EPOLL_CTL_ADD, *it, &ev);
       if (UNLIKELY (ret < 0))
-        return crun_make_error (err, errno, "epoll_ctl add '%d'", *it);
+        return crun_make_error (err, errno, "epoll_ctl add `%d`", *it);
     }
 
   ret = epollfd;
@@ -1646,11 +1666,11 @@ mark_or_close_fds_ge_than (int n, bool close_now, libcrun_error_t *err)
   if (ret == 0)
     return 0;
   if (ret < 0 && errno != EINVAL && errno != ENOSYS && errno != EPERM)
-    return crun_make_error (err, errno, "close_range from %d", n);
+    return crun_make_error (err, errno, "close_range from `%d`", n);
 
   cfd = open ("/proc/self/fd", O_DIRECTORY | O_RDONLY | O_CLOEXEC);
   if (UNLIKELY (cfd < 0))
-    return crun_make_error (err, errno, "open /proc/self/fd");
+    return crun_make_error (err, errno, "open `/proc/self/fd`");
 
   ret = fstatfs (cfd, &sfs);
   if (UNLIKELY (ret < 0))
@@ -1661,7 +1681,7 @@ mark_or_close_fds_ge_than (int n, bool close_now, libcrun_error_t *err)
 
   dir = fdopendir (cfd);
   if (UNLIKELY (dir == NULL))
-    return crun_make_error (err, errno, "cannot fdopendir /proc/self/fd");
+    return crun_make_error (err, errno, "cannot fdopendir `/proc/self/fd`");
 
   /* Now it is owned by dir.  */
   cfd = -1;
@@ -2079,19 +2099,9 @@ copy_recursive_fd_to_fd (int srcdirfd, int dfd, const char *srcname, const char
       ret = fchmodat (destdirfd, de->d_name, mode & ALLPERMS, AT_SYMLINK_NOFOLLOW);
       if (UNLIKELY (ret < 0))
         {
+          /* If the operation fails with ENOTSUP we are dealing with a symlink, so ignore it.  */
           if (errno == ENOTSUP)
-            {
-              proc_fd_path_t proc_path;
-              cleanup_close int fd = -1;
-
-              fd = openat (destdirfd, de->d_name, O_PATH | O_NOFOLLOW);
-              if (UNLIKELY (fd < 0))
-                return crun_make_error (err, errno, "open `%s/%s`", destname, de->d_name);
-
-              get_proc_self_fd_path (proc_path, fd);
-
-              ret = chmod (proc_path, mode & ALLPERMS);
-            }
+            continue;
 
           if (UNLIKELY (ret < 0))
             return crun_make_error (err, errno, "chmod `%s/%s`", destname, de->d_name);
